"""Command Line  Interface for Crewcal, a tool that extracts flight data from an airline crew schedule."""

import pathlib

import click
from halo import Halo

from crewcal import schedule
from crewcal.llm_extract import OpenAISchedule


@click.group()
def cli():
    """Crewcal is a tool that extracts flight data from an airline crew schedule.

    An LLM (Large Language Model) is used to extract data from the unstructured schedule
    (in pdf format). This data is then converted to iCalendar format (or alternatively
    a crewcal specific json file) and saved to a specified file.

    Note that the an environment variable named "OPENAI_API_KEY" must be set
    with your OpenAI API key. At present (november 2023) each 'extract' costs just
    under USD 0.01 (charged to your OpenAI account).
    """


@click.command
@click.option(
    "--overwrite",
    "-o",
    is_flag=True,
    help="Overwrite the iCalendar file if it already exists.",
)
@click.argument("json")
@click.argument("ical")
def convert(json: str, ical: str, overwrite: bool) -> int:
    """Convert crewcal json schedule file to iCalendar file.

    The JSON file should be in crewcal json format.

    \b
    Args:
        JSON (str): Path to json file containing schedule.
        ICAL (str): Path to iCalendar file.
    """  # noqa: D301
    json_path = pathlib.Path(json)
    ical_path = pathlib.Path(ical)

    if not ical_path.suffix:
        ical_path = ical_path.with_suffix(".ics")

    if ical_path.is_file() and not overwrite:
        click.echo(
            f"File {ical_path} already exists. Consider using '--overwrite' option."
        )
        return -1

    if not json_path.is_file() and not json_path.with_suffix(".json").is_file():
        click.echo(f"User specified file '{json_path}' not found.")
        return -1

    if not json_path.is_file():
        json_path_modified = json_path.with_suffix(".json")
        click.echo(
            f"File '{json_path}' not found. Using '{json_path_modified}' instead."
        )
        json_path = json_path_modified

    sched = schedule.Schedule.from_json(str(json_path))
    sched.to_icalendar_file(str(ical_path))

    return 0


@click.command
@click.option(
    "--to-json",
    "-j",
    is_flag=True,
    help="Save to crewcal json schedule file (instead of Icalendar).",
)
@click.option(
    "--overwrite",
    "-o",
    is_flag=True,
    help="Overwrite the target file if it already exists.",
)
@click.argument("sourcefile")
@click.argument("targetfile")
def extract(sourcefile: str, targetfile: str, to_json: bool, overwrite: bool) -> int:
    """Extract schedule from pdf file and save to iCalendar format (or json).

    The saved json is in a format specific to crewcal. If saved to json,
    use 'crewcal convert' to convert it to iCalendar format.

    \b
    Args:
        SOURCEFILE (str): Path to pdf file.
        TARGETFILE (str): Path to iCalendar (or json) file.
    """  # noqa: D301
    out_path = pathlib.Path(targetfile)
    source_path = pathlib.Path(sourcefile)

    if not (out_path.suffix):
        out_path = out_path.with_suffix(".json" if to_json else ".ics")

    if out_path.is_file() and not overwrite:
        click.echo(
            f"File '{out_path}' already exists. Consider using '--overwrite' option."
        )
        return -1

    if not source_path.is_file() and not source_path.with_suffix(".pdf").is_file():
        click.echo(f"User specified file '{source_path}' not found.")
        return -1

    if not source_path.is_file():
        source_path_modified = source_path.with_suffix(".pdf")
        click.echo(
            f"File '{source_path}' not found. Using '{source_path_modified}' instead."
        )
        source_path = source_path_modified

    with (
        Halo(text="Extracting schedule, saving to iCalendar format.", spinner="dots")
        if not to_json
        else Halo(
            text="Extracting schedule, saving to crewcal json format.", spinner="dots"
        )
    ) as spinner:
        (
            OpenAISchedule(
                schedule_path=str(source_path), to_icalendar_file=str(out_path)
            )
            if not to_json
            else OpenAISchedule(
                schedule_path=str(source_path), to_json_file=str(out_path)
            )
        )
        spinner.info(f"Extracted schedule saved to {out_path}.")

    return 0


@click.command
@click.option(
    "--overwrite",
    "-o",
    is_flag=True,
    help="Overwrite the file if file with same name already exists in target folder.",
)
@click.argument("sourcefile")
@click.argument("targetfolder")
def hotels(sourcefile: str, targetfolder: str, overwrite: bool) -> int:
    """Extract hotel contact information from pdf schedule.

    A vCard file will be saved in the target folder for each hotel.

    \b
    Args:
        SOURCEFILE (str): Path to pdf file.
        TARGETFOLDER (str): Path to folder where contact files will be saved.
    """  # noqa: D301
    out_path = pathlib.Path(targetfolder)
    source_path = pathlib.Path(sourcefile)

    # TODO: Ensure out_path is a folder
    # TODO: Create folder if not exists

    if not source_path.is_file() and not source_path.with_suffix(".pdf").is_file():
        click.echo(f"User specified file '{source_path}' not found.")
        return -1

    """ if not out_path.is_dir():
        click.echo(f"User specified target '{targetfolder}' is not a folder.")
        return -1 """

    if not source_path.is_file():
        source_path_modified = source_path.with_suffix(".pdf")
        click.echo(
            f"File '{source_path}' not found. Using '{source_path_modified}' instead."
        )
        source_path = source_path_modified

    with Halo(
        text="Extracting hotel contacts, saving to vCard format.", spinner="dots"
    ) as spinner:
        (OpenAISchedule(schedule_path=str(source_path), to_hotel_folder=str(out_path)))
        spinner.info(f"Extracted hotel information saved to {out_path}.")
    return 0


cli.add_command(extract)
cli.add_command(convert)
cli.add_command(hotels)

if __name__ == "__main__":
    cli()



"""Prompts for the LLM."""

template_flight_schedule = """
Your task is to convert the information I will provide into a list of calendar events.

The information contains a list of duties for an airline crew member. Duties consist of one or more flights. You will represent each duty as a calendar event.

The section with Hotel Information contains no new events. Instead it contains the hotel at destination for some of the flights.

Each event contains the following items:
- Departure date in yyyy-mm-dd format
- Departure time in HH:mm format
- Duties which contains one or multiple flight numbers
- Summary which contains the origin and departure airports
- Origin airport code. Sometimes multiple flights exist. Capture all origin airports.
- Origin airport name
- Origin timezone in ISO 8601 format (for example 'America/Toronto')
- Destination airport code
- Destination airport name
- Destination timezone
- Arrival date. A '+1' means the next day. Only capture yyyy-mm-dd format.
- Arrival time, the last occurrence of a time for each day. A '+1' means the next day. Only capture HH:mm format.
- List of crew members
- A list of all times found. Only capture HH:mm format.
- List of all airport codes.
- Hotel information at destination. Obtain this from the document section 'Hotel Information'. It includes the name, full address and phone number of the hotel at destination. For some flights this does not exist. In that case keep this part empty.

Always include all items in your output even if they are empty.
"""


template_hotel_contacts = """
The provided document may contain a section with hotel information, base your answer on this section.

The hotel section may contain information about one or more hotels.
For each hotel, provide:
- a suggested file name for the vcf file (vCard format).
- the contact information of the hotel, in vCard format.

If you cannot find any hotel contact information, return an empty list.
"""



"""Class(es) defining contact information for hotels

"""

from typing import List

from pydantic import BaseModel, Field


class Hotel(BaseModel):
    """Contact details of a hotel on a flight schedule for an airline crew member."""

    vcf_file_name: str = Field(
        description="Proposed file name for the vCard file, including file extension."
    )
    hotel_contact: str = Field(
        description='Full contact information in vCard format. Phone numbers use the "+ country code" format; do not include "tel" in the phone number. Country is always included. Always include an N record (a structured representation of the name) and an FN record. Include the hotel name in the ORG record. Always include the categories: hotel, crew. Ensure it is compatible with iOS, gmail, Android, Windows and MAC.'
    )


class Hotels(BaseModel):
    """List of hotel information."""

    hotels: List[Hotel]



"""Classes defining a flight schedule for an airline crew member.

The schedule is a list of events, each of which contains the details of a set of flight.

"""

import json
import logging
from pathlib import Path
from typing import List

import ics
import pendulum
from pydantic import BaseModel, Field, TypeAdapter


class Event(BaseModel):
    """The details of a flight on a flight schedule for an airline crew member."""

    starting_date: str
    starting_time: str
    duties: List[str]
    summary: str
    description: str
    departure_airport: List[str]
    departure_airport_name: List[str]
    departure_timezone: List[str] = Field(description='for example "America/Toronto"')
    destination_airport: List[str]
    destination_airport_name: List[str]
    destination_timezone: List[str]
    end_date: str
    end_time: str
    crew_list: List[str]
    list_times: List[str]
    list_airport_codes: List[str]
    hotel_information: str

    def list_airport_pairs(self):
        """Generate a formatted string of airport pairs."""
        airport_pairs = zip(
            self.duties, self.departure_airport_name, self.destination_airport_name
        )
        formatted_pairs = (
            f"{flight_num}: {dep} - {dest}" for flight_num, dep, dest in airport_pairs
        )
        return "\n".join(formatted_pairs)

    def get_description(self):
        """Generate a formatted string of flight event details.

        Returns:
            str: description
        """
        return "\n".join(
            [
                self.list_airport_pairs(),
                *self.crew_list,
                f"\nFlight ends in {self.destination_timezone[-1]} timezone.",
                self.hotel_information,
            ]
        )

    def get_begin(self):
        """Get the starting time of the flight event.

        Returns:
            datetime: start datetime
        """
        return pendulum.from_format(
            f"{self.starting_date} {self.starting_time}",
            "YYYY-MM-DD HH:mm",
            tz=self.departure_timezone[0],
        )

    def get_end(self):
        """Get the ending time of the flight event.

        Returns:
            datetime: end datetime
        """
        return pendulum.from_format(
            f"{self.end_date} {self.list_times[-1]}",
            "YYYY-MM-DD HH:mm",
            tz=self.destination_timezone[-1],
        )


class Schedule(BaseModel):
    """A flight schedule for an airline crew member."""

    events: List[Event]

    def to_icalendar(self):
        """Generates a iCalendar representation of the schedule.

        Returns:
            ics.Calendar: The iCalendar object containing the events.
        """
        calendar = ics.Calendar(creator="Ternyx - crewcal")

        events = [
            ics.Event(
                name=flight.summary,
                description=flight.get_description(),
                begin=flight.get_begin(),
                end=flight.get_end(),
            )
            for flight in self.events
        ]

        calendar.events.update(events)

        return calendar

    def to_icalendar_file(self, filename: str) -> None:
        """Write the iCalendar representation of the schedule to a file.

        Parameters:
            filename (str): The name of the file to write the iCalendar data to.
        """
        try:
            with Path(filename).open("w") as f:
                f.write(self.to_icalendar().serialize())
        except Exception as e:
            logging.warning(f"Error writing to file: {e}")

    def json_dumps(self, indent=2):
        """Show the schedule as a JSON string.

        Parameters:
            indent (int):

        Returns:
            str: The JSON representation of the schedule.
        """
        return json.dumps(self.model_dump(mode="json"), indent=indent)

    @staticmethod
    def from_json(filename: str = "./etc/sched.json") -> "Schedule":
        """Load flight events from a JSON file and create a schedule.

        Args:
            filename (str): The path of the JSON file.

        Returns:
            Schedule: A Schedule object containing the schedule of flights.
        """
        with Path(filename).open("r") as openfile:
            events_from_file = json.load(openfile)

        adapter = TypeAdapter(List[Event])
        events = adapter.validate_python(events_from_file)
        return Schedule(events=events)

    @staticmethod
    def from_json_string(json_string: str) -> "Schedule":
        """Load flight events from a JSON string and create a schedule.

        Args:
            json_string (str): The events in JSON format.

        Returns:
            Schedule: A Schedule object containing the schedule of flights.
        """
        adapter = TypeAdapter(List[Event])
        events = adapter.validate_python(json_string)
        return Schedule(events=events)



"""Entry point for crewcal package command line interface.

Convert an airline crew schedule pdf into iCalendar format.
"""

import logging

from crewcal import schedule


def main():
    """_summary_."""
    sched = schedule.Schedule.from_json("sched.json")

    logging.warning(sched.to_icalendar().serialize())

    sched.to_icalendar_file("dien.ics")

    logging.warning(sched.json_dumps())

    return -1


if __name__ == "__main__":
    main()



"""Read an airline crew schedule in pdf format.

Read the schedule from a pdf file using LLM with OpenAI
to extract flight details in a structured format.

As at nov 2023 about 0.75 US cents per call in OpenAI API costs.
"""

import json
import logging
import os
from pathlib import Path

import openai
from dotenv import find_dotenv, load_dotenv
from langchain.chat_models import ChatOpenAI
from langchain.document_loaders import PyPDFLoader
from langchain.output_parsers import PydanticOutputParser
from langchain.output_parsers.openai_functions import JsonKeyOutputFunctionsParser
from langchain.prompts import ChatPromptTemplate
from langchain.schema.runnable import RunnablePassthrough
from langchain.utils.openai_functions import convert_pydantic_to_openai_function

from crewcal.hotel import Hotels
from crewcal.llm_prompts import template_flight_schedule, template_hotel_contacts
from crewcal.schedule import Schedule

_ = load_dotenv(find_dotenv())
try:
    openai.api_key = os.environ["OPENAI_API_KEY"]
except KeyError:
    logging.warning("WARNING - Environment variable OPENAI_API_KEY should be set.")


class OpenAISchedule:
    """Extracts flight data from an airline crew schedule provided in PDF format.

    Uses an LLM for extraction. Exctracted data is in a structured format.
    """

    schedule_path: str
    extracted_schedule: str = ""
    extracted_hotels = []
    llm_model_name: str = "gpt-4o-mini-2024-07-18"

    def __init__(
        self,
        schedule_path: str,
        to_json_file: str = "",
        to_icalendar_file: str = "",
        to_hotel_folder: str = "",
    ) -> None:
        """Sets up the object using the provided schedule_path. Additionally, it allows for an optional to_file path where the schedule can be extracted.

        If the to_file path isn't specified during object creation, you'll need to call the extraction method explicitly at a later time.

        Sample uses:
        - sched = OpenAISchedule("schedule.pdf", to_json_file="schedule.json") - to extract data from a PDF file and save it to a JSON file
        - sched.write_icalendar("schedule.ics") - to save he previously extract data to an iCalendar file
        - sched.write_json("schedule.json") - to save the extracted data to a JSON file (note that this already happened in the first example)

        - sched = OpenAISchedule.from_json("schedule.json") - create object from JSON. Note that extract() has no functionality and schedule path is empty.
        - sched.write_icalendar("schedule.ics")

        - sched = OpenAISchedule("schedule.pdf")
        - sched.extract()
        - sched.write_icalendar("schedule.ics")

        - sched = OpenAISchedule("schedule.pdf", to_icalendar_file="schedule.ics") - to extract data from a PDF file and save it to a JSON file

        Parameters:
            schedule_path (str): The path to the schedule PDF file.
            to_json_file (str, optional): The path to the file where the schedule will be extracted.
            to_icalendar_file (str, optional): The path to the file where the iCalendar representation of the schedule will be saved.
            to_hotel_folder (str, optional): The folder where hotel contact info cards will be saved.

        Returns:
            None
        """
        self.schedule_path = schedule_path

        if to_json_file:
            self.extract(to_json_file)

        if to_icalendar_file:
            if not self.extracted_schedule:
                self.extract()
            self.write_icalendar(to_icalendar_file)

        if to_hotel_folder:
            self.extract_hotels(Path(to_hotel_folder))

    def extract(self, to_file: str = "") -> None:
        """Uses LLM to extract event data from a schedule PDF file and optionally saves it to a JSON file.

        Parameters:
            to_file (str, optional): The path to the output JSON file. If not provided, the extracted data will not be saved.

        Returns:
            None
        """
        full_sched_doc = self.read_schedule_pdf(self.schedule_path)

        if full_sched_doc:
            model = ChatOpenAI(model_name=self.llm_model_name, temperature=0)
            prompt = ChatPromptTemplate.from_messages(
                [("system", template_flight_schedule), ("human", "{input}")]
            )

            schedule_extraction_function = [
                convert_pydantic_to_openai_function(Schedule)
            ]

            extraction_model = model.bind(
                functions=schedule_extraction_function,
                function_call={"name": "Schedule"},
            )

            extraction_chain = (
                prompt
                | extraction_model
                | JsonKeyOutputFunctionsParser(key_name="events")
            )

            logging.warning(
                "WARNING - This script costs ~0.75 US cents per call in OpenAI API costs (GPT-3.5)."
            )

            from langchain.callbacks import get_openai_callback

            with get_openai_callback() as cb:
                self.extracted_schedule = extraction_chain.invoke(
                    {"input": full_sched_doc}
                )

            logging.warning("Actual OpenAI API cost in USD:" + str(cb.total_cost))

        if to_file:
            self.write_json(to_file)

    def extract_hotels(self, to_folder: Path) -> None:
        """Uses an LLM to extract hotel contact information from a flight schedule.

        This is converted into vCard format by the same LLM. The results are saved into
        vCard files.

        Args:
            to_folder (Path): Destination folder of the vCard files.

        Returns:
            None
        """
        full_sched_doc = self.read_schedule_pdf(self.schedule_path)

        if full_sched_doc:
            model = ChatOpenAI(model_name=self.llm_model_name, temperature=0)
            hotel_parser = template_hotel_contacts
            hotel_parser = PydanticOutputParser(pydantic_object=Hotels)
            prompt = ChatPromptTemplate.from_template(
                template_hotel_contacts + "\n\n"
                "Document: {document}\n\n"
                "{format_instructions}"
            )

            extraction_chain = (
                {
                    "document": RunnablePassthrough(),
                    "format_instructions": lambda _: hotel_parser.get_format_instructions(),
                }
                | prompt
                | model
                | hotel_parser
            )

            logging.warning(
                "WARNING - This script costs ~0.75 US cents per call in OpenAI API costs (GPT-3.5)."
            )

            from langchain.callbacks import get_openai_callback

            with get_openai_callback() as cb:
                self.extracted_hotels = extraction_chain.invoke(full_sched_doc)
                if str(cb.total_cost) != "":
                    logging.warning(
                        "Actual OpenAI API cost in USD:" + str(cb.total_cost)
                    )

            if not to_folder.exists() and len(self.extracted_hotels.hotels) > 0:
                to_folder.mkdir(parents=True, exist_ok=True)

            for hotel in self.extracted_hotels.hotels:
                destination_file = to_folder / hotel.vcf_file_name
                with open(destination_file, "w") as file:
                    file.write(hotel.hotel_contact)

    def read_schedule_pdf(self, filepath: str = "") -> str:
        """Reads the contents of a schedule PDF file and returns as a document for an LLM input.

        Args:
            filepath (str): The path to the PDF file.

        Returns:
            str: The concatenated text of all pages in the PDF file.
        """
        if filepath:
            loader = PyPDFLoader(str(filepath))
            documents = loader.load()
            full_sched_doc = "".join([doc.page_content for doc in documents])
        else:
            full_sched_doc = ""

        return full_sched_doc

    def json_dumps(self) -> str:
        """Returns the JSON representation of the schedule in a basic formatted string.

        Returns:
            str: The JSON representation of the schedule.
        """
        return json.dumps(self.extracted_schedule, indent=2, ensure_ascii=False)

    def write_json(self, filepath: str = "./sched.json") -> None:
        """Writes the JSON representation of the schedule to a file."""
        with Path(filepath).open("w") as outfile:
            json.dump(self.extracted_schedule, outfile)

    def write_icalendar(self, filepath: str) -> None:
        """Writes the iCalendar representation of the schedule to a file."""
        sched = Schedule.from_json_string(self.extracted_schedule)
        sched.to_icalendar_file(filepath)
        del sched

    @staticmethod
    def from_json(filepath: str) -> "OpenAISchedule":
        """Initializes a new OpenAISchedule object from a JSON file.

        Args:
            filepath (str): The path to the JSON file.

        Returns:
            OpenAISchedule: The initialized OpenAISchedule object.
        """
        with Path(filepath).open("r") as openfile:
            extracted_schedule = json.load(openfile)

        new_schedule = OpenAISchedule("")
        new_schedule.extracted_schedule = extracted_schedule
        return new_schedule


# TODO: Crew member names are captured, but not seniority and function.
# TODO: Must be asynchronous somehow as OpenAI may take a while to respond.
